import asyncio
import logging
from datetime import datetime, timezone
from threading import Thread
from typing import Any

import httpx
from asgiref.sync import sync_to_async
from django.db.models import QuerySet, signals
from django.db.models.base import ModelBase
from django.dispatch import receiver

from .settings import webhook_settings
from .typing import (
    TYPE_CHECKING,
    Callable,
    ClientMethodKwargs,
    Dict,
    HooksData,
    JSONData,
    Optional,
    PostDeleteData,
    PostSaveData,
    Set,
    SignalChoices,
)
from .utils import get_webhookhook_model, reference_for_model, tasks_as_completed, truncate


if TYPE_CHECKING:
    from .models import Webhook


__all__ = [
    "default_error_handler",
    "default_post_delete_handler",
    "default_post_save_handler",
    "sync_task_handler",
    "thread_task_handler",
    "webhook_delete_handler",
    "webhook_update_create_handler",
]


logger = logging.getLogger(__name__)


@receiver(signals.post_save, dispatch_uid=webhook_settings.DISPATCH_UID_POST_SAVE)
def webhook_update_create_handler(sender: ModelBase, **kwargs) -> None:  # pylint: disable=unused-argument
    kwargs: PostSaveData
    ref = reference_for_model(type(kwargs["instance"]))

    hooks: Optional[HooksData] = webhook_settings.HOOKS.get(ref)
    if hooks is None:
        return
    if hooks is ...:
        webhook_settings.TASK_HANDLER(
            default_post_save_handler,
            ref=ref,
            data=webhook_settings.SERIALIZER(kwargs["instance"]),
            created=kwargs["created"],
        )
        return

    hook = hooks.get("CREATE") if kwargs["created"] else hooks.get("UPDATE")

    if hook is None:
        return
    if hook is ...:
        webhook_settings.TASK_HANDLER(
            default_post_save_handler,
            ref=ref,
            data=webhook_settings.SERIALIZER(kwargs["instance"]),
            created=kwargs["created"],
        )
        return

    webhook_settings.TASK_HANDLER(
        hook,
        ref=ref,
        data=webhook_settings.SERIALIZER(kwargs["instance"]),
        created=kwargs["created"],
    )


@receiver(signals.post_delete, dispatch_uid=webhook_settings.DISPATCH_UID_POST_DELETE)
def webhook_delete_handler(sender: ModelBase, **kwargs) -> None:  # pylint: disable=unused-argument
    kwargs: PostDeleteData
    ref = reference_for_model(type(kwargs["instance"]))

    hooks: Optional[HooksData] = webhook_settings.HOOKS.get(ref)
    if hooks is None:
        return
    if hooks is ...:
        webhook_settings.TASK_HANDLER(
            default_post_delete_handler,
            ref=ref,
            data=webhook_settings.SERIALIZER(kwargs["instance"]),
        )
        return

    hook = hooks.get("DELETE")

    if hook is None:
        return
    if hook is ...:
        webhook_settings.TASK_HANDLER(
            default_post_delete_handler,
            ref=ref,
            data=webhook_settings.SERIALIZER(kwargs["instance"]),
        )
        return

    webhook_settings.TASK_HANDLER(
        hook,
        ref=ref,
        data=webhook_settings.SERIALIZER(kwargs["instance"]),
    )


def default_error_handler(hook: "Webhook", error: Optional[Exception]) -> None:  # pylint: disable=unused-argument
    """Default handler for errors from webhooks.

    :param hook: Hook that failed
    :param error: Exception if one was caused due to, e.g., timeout.
    """


def thread_task_handler(hook: Callable[..., None], **kwargs: Any) -> None:
    thread = Thread(target=hook, kwargs=kwargs)
    thread.start()


def sync_task_handler(hook: Callable[..., None], **kwargs: Any) -> None:
    hook(**kwargs)


def default_post_save_handler(ref: str, data: JSONData, created: bool) -> None:
    webhook_model = get_webhookhook_model()

    signal_types = SignalChoices.create_choises() if created else SignalChoices.update_choises()
    hooks = webhook_model.objects.get_for_ref(ref, signals=signal_types)

    if not hooks.exists():
        return

    method_data: Dict[int, ClientMethodKwargs] = {}
    for hook in hooks:
        method_data[hook.id] = webhook_settings.CLIENT_KWARGS(hook)
        method_data[hook.id].setdefault("headers", {})
        method_data[hook.id]["headers"].update(hook.default_headers())
        method_data[hook.id]["json"] = data

    asyncio.run(fire_webhooks(hooks, method_data))


def default_post_delete_handler(ref: str, data: JSONData) -> None:
    webhook_model = get_webhookhook_model()

    signal_types = SignalChoices.delete_choises()
    hooks = webhook_model.objects.get_for_ref(ref, signals=signal_types)

    if not hooks.exists():
        return

    method_data: Dict[int, ClientMethodKwargs] = {}
    for hook in hooks:
        method_data[hook.id] = webhook_settings.CLIENT_KWARGS(hook)
        method_data[hook.id].setdefault("headers", {})
        method_data[hook.id]["headers"].update(hook.default_headers())
        method_data[hook.id]["json"] = data

    asyncio.run(fire_webhooks(hooks, method_data))


async def fire_webhooks(hooks: QuerySet["Webhook"], method_data: Dict[int, ClientMethodKwargs]) -> None:
    futures: Set[asyncio.Task] = set()
    hooks_by_name: Dict[str, "Webhook"] = {hook.name: hook for hook in hooks}
    webhook_model = get_webhookhook_model()

    async with httpx.AsyncClient(timeout=webhook_settings.TIMEOUT, follow_redirects=True) as client:
        for hook in hooks:
            futures.add(asyncio.Task(client.post(hook.endpoint, **method_data[hook.id]), name=hook.name))

        async for task in tasks_as_completed(futures):
            hook = hooks_by_name[task.get_name()]

            try:
                response: httpx.Response = task.result()
            except Exception as error:  # pylint: disable=broad-except
                logger.exception(f"Webhook {hook.name!r} failed.", exc_info=error)
                hook.last_failure = datetime.now(tz=timezone.utc)
                webhook_settings.ERROR_HANDLER(hook, error)
                continue

            if response.status_code // 100 == 2:
                hook.last_success = datetime.now(tz=timezone.utc)
                if hook.keep_last_response:
                    hook.last_response = truncate(response.content.decode())

            else:
                hook.last_failure = datetime.now(tz=timezone.utc)
                webhook_settings.ERROR_HANDLER(hook, None)
                if hook.keep_last_response:
                    hook.last_response = truncate(response.content.decode())

    await sync_to_async(webhook_model.objects.bulk_update)(
        objs=hooks,
        fields=[
            "last_success",
            "last_failure",
            "last_response",
        ],
    )



import logging
from base64 import b64decode, b64encode
from os import urandom

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from django.core.exceptions import ValidationError
from django.db import models

from .utils import decode_cipher_key


__all__ = [
    "TokenField",
]


logger = logging.getLogger(__name__)


class TokenField(models.CharField):
    """Encrypt token with a cipher before saving it."""

    def from_db_value(self, value: str, *args, **kwargs) -> str:
        if not value:
            return value

        key = decode_cipher_key()

        string = b64decode(value)
        nonce = string[:12]
        data = string[12:]
        cipher = AESGCM(key=key)

        try:
            decrypted_token = cipher.decrypt(nonce, data, None)
        except InvalidTag as error:
            raise ValidationError("Wrong cipher key.") from error

        return decrypted_token.decode()

    def get_prep_value(self, value: str, *args, **kwargs) -> str:
        if not value:
            return value

        key = decode_cipher_key()

        nonce = urandom(12)
        cipher = AESGCM(key)
        encrypted_token = cipher.encrypt(nonce, value.encode(encoding="utf-8"), None)
        encoded_token = b64encode(nonce + encrypted_token).decode()
        return encoded_token



import asyncio
import logging
from base64 import b64decode, b64encode
from functools import lru_cache
from importlib import import_module
from os import urandom
from sys import modules

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.models import Model
from django.db.models.base import ModelBase

from .serializers import webhook_serializer
from .settings import webhook_settings
from .typing import TYPE_CHECKING, Any, ClientMethodKwargs, Generator, JSONData, Literal, Set, Type


if TYPE_CHECKING:
    from .models import Webhook, WebhookBase


__all__ = [
    "decode_cipher_key",
    "default_client_kwargs",
    "default_serializer",
    "get_webhookhook_model",
    "is_dict",
    "model_from_reference",
    "random_cipher_key",
    "reference_for_model",
    "tasks_as_completed",
    "truncate",
]


logger = logging.getLogger(__name__)


def is_dict(value: str) -> None:
    if not isinstance(value, dict):
        raise ValidationError("Headers should always be a dict.")


def truncate(string: str, limit: int = 10_000) -> str:
    if len(string) > limit:
        string = string[: limit - 3] + "..."
    return string


def random_cipher_key(length: Literal[16, 24, 32] = 16) -> str:
    return b64encode(urandom(length)).decode()


# 'value' as parameter so that can be used as a field validator
def decode_cipher_key(value: str = "") -> bytes:  # pylint: disable=unused-argument
    try:
        return b64decode(webhook_settings.CIPHER_KEY)
    except TypeError as error:
        raise ValidationError("Cipher key not set.") from error
    except Exception as error:  # pylint: disable=broad-except
        raise ValidationError("Invalid cipher key.") from error


def default_serializer(instance: Model) -> JSONData:
    if hasattr(instance, "webhook_data") and callable(instance.webhook_data):
        return instance.webhook_data()

    return webhook_serializer.serialize([instance])


def default_client_kwargs(hook: "Webhook") -> ClientMethodKwargs:  # pylint: disable=unused-argument
    return ClientMethodKwargs()


@lru_cache(maxsize=None)
def get_webhookhook_model() -> Type["WebhookBase"]:
    ref = getattr(settings, "SIGNAL_WEBHOOKS_CUSTOM_MODEL", "signal_webhooks.models.Webhook")

    if ref == "signal_webhooks.models.Webhook":
        from .models import Webhook  # pylint: disable=import-outside-toplevel

        return Webhook

    try:
        model = model_from_reference(ref, check_hooks=False)
    except ValidationError as error:
        raise ImproperlyConfigured(f"{ref!r} is not a model that can be imported.") from error

    from .models import WebhookBase  # pylint: disable=import-outside-toplevel

    if not issubclass(model, WebhookBase):
        base_ref = f"{WebhookBase.__module__}.{WebhookBase.__name__}"
        raise ImproperlyConfigured(f"{ref!r} is not a subclass of a {base_ref!r}.")

    return model


def model_from_reference(ref: str, check_hooks: bool = True) -> ModelBase:
    msg = f"Could not import {ref!r}"
    try:
        module_path, class_name = ref.rsplit(".", 1)
    except ValueError as error:
        raise ValidationError(f"{msg}. {ref!r} doesn't look like a module path.") from error

    if module_path not in modules or (
        # Module is not fully initialized.
        getattr(modules[module_path], "__spec__", None) is not None
        and getattr(modules[module_path].__spec__, "_initializing", False) is True
    ):  # pragma: no cover
        try:
            import_module(module_path)
        except ImportError as error:
            raise ValidationError(f"{msg}. {error.__class__.__name__!r}: {error}.") from error

    try:
        model_type = getattr(modules[module_path], class_name)
    except AttributeError as error:
        raise ValidationError(f"{msg}. Module {module_path!r} does not define {class_name!r}.") from error

    if not isinstance(model_type, ModelBase):
        raise ValidationError(f"{ref!r} is not a django model.")

    if check_hooks and ref not in webhook_settings.HOOKS:
        raise ValidationError(f"Webhooks not defined for {ref!r}.")

    return model_type


def reference_for_model(model: ModelBase) -> str:
    return f"{model.__module__}.{model.__name__}"


async def tasks_as_completed(tasks: Set[asyncio.Task]) -> Generator[asyncio.Task, Any, None]:
    done = asyncio.Queue()

    def _on_completion(task_: asyncio.Task) -> None:
        tasks.remove(task_)
        done.put_nowait(task_)

    async def _wait_for_one() -> asyncio.Task:
        return await done.get()

    for task in tasks:
        task.add_done_callback(_on_completion)

    for _ in range(len(tasks)):
        yield await _wait_for_one()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['signal_webhooks', 'signal_webhooks.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.1',
 'asgiref>=3.5.0',
 'cryptography>=36.0.0',
 'django-settings-holder>=0.0.4',
 'httpx>=0.23.0']

setup_kwargs = {
    'name': 'django-signal-webhooks',
    'version': '0.2.0',
    'description': 'Add webhooks to django using signals.',
    'long_description': '# Django Signal Webhooks\n\n[![Coverage Status][coverage-badge]][coverage]\n[![GitHub Workflow Status][status-badge]][status]\n[![PyPI][pypi-badge]][pypi]\n[![GitHub][licence-badge]][licence]\n[![GitHub Last Commit][repo-badge]][repo]\n[![GitHub Issues][issues-badge]][issues]\n[![Downloads][downloads-badge]][pypi]\n\n[![Python Version][version-badge]][pypi]\n[![Django Version][django-badge]][pypi]\n\n```shell\npip install django-signal-webhooks\n```\n\n---\n\n**Documentation**: [https://mrthearman.github.io/django-signal-webhooks/](https://mrthearman.github.io/django-signal-webhooks/)\n\n**Source Code**: [https://github.com/MrThearMan/django-signal-webhooks/](https://github.com/MrThearMan/django-signal-webhooks/)\n\n---\n\nThis library enables you to add webhooks to a Django project for any create/update/delete\nevents on your models with a simple configuration. New webhooks can be added in the\nadmin panel, with or without authentication, with plenty of hooks into the webhook sending\nprocess to customize them for your needs.\n\n```python\n# project/settings.py\n\n# Add to instaled apps\nINSTALLED_APPS = [\n    ...\n    "signal_webhooks",\n    ...\n]\n\n# Add default webhook configuration to the User model\nSIGNAL_WEBHOOKS = {\n    "HOOKS": {\n        "django.contrib.auth.models.User": ...,\n    },\n}\n```\n\n[coverage-badge]: https://coveralls.io/repos/github/MrThearMan/django-signal-webhooks/badge.svg?branch=main\n[status-badge]: https://img.shields.io/github/workflow/status/MrThearMan/django-signal-webhooks/Test\n[pypi-badge]: https://img.shields.io/pypi/v/django-signal-webhooks\n[licence-badge]: https://img.shields.io/github/license/MrThearMan/django-signal-webhooks\n[repo-badge]: https://img.shields.io/github/last-commit/MrThearMan/django-signal-webhooks\n[issues-badge]: https://img.shields.io/github/issues-raw/MrThearMan/django-signal-webhooks\n[version-badge]: https://img.shields.io/pypi/pyversions/django-signal-webhooks\n[downloads-badge]: https://img.shields.io/pypi/dm/django-signal-webhooks\n[django-badge]: https://img.shields.io/pypi/djversions/django-signal-webhooks\n\n[coverage]: https://coveralls.io/github/MrThearMan/django-signal-webhooks?branch=main\n[status]: https://github.com/MrThearMan/django-signal-webhooks/actions/workflows/test.yml\n[pypi]: https://pypi.org/project/django-signal-webhooks\n[licence]: https://github.com/MrThearMan/django-signal-webhooks/blob/main/LICENSE\n[repo]: https://github.com/MrThearMan/django-signal-webhooks/commits/main\n[issues]: https://github.com/MrThearMan/django-signal-webhooks/issues\n',
    'author': 'Matti Lamppu',
    'author_email': 'lamppu.matti.akseli@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://mrthearman.github.io/django-signal-webhooks/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    FrozenSet,
    Generator,
    Iterator,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypedDict,
    Union,
)

import django
from django.db import models
from django.db.models import QuerySet
from django.db.models.base import Model
from django.db.models.signals import ModelSignal
from httpx._client import UseClientDefault
from httpx._types import (
    AuthTypes,
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestFiles,
    TimeoutTypes,
)


__all__ = [
    "Any",
    "Callable",
    "ClientMethodKwargs",
    "Coroutine",
    "Dict",
    "Generator",
    "HooksData",
    "Iterator",
    "JSONData",
    "JSONValue",
    "List",
    "Literal",
    "NamedTuple",
    "Optional",
    "PostDeleteData",
    "PostSaveData",
    "Set",
    "SignalChoices",
    "Tuple",
    "Type",
    "TYPE_CHECKING",
    "TypedDict",
    "Union",
    "Sequence",
]


class PostSaveData(TypedDict):
    signal: ModelSignal
    instance: Model
    created: bool
    raw: bool
    using: str  # database name
    update_fields: Optional[FrozenSet[str]]


class PostDeleteData(TypedDict):
    signal: ModelSignal
    instance: Model
    using: str  # database name

    if django.VERSION >= (4, 1):
        origin: Union[Model, QuerySet]


JSONValue = Union[str, int, float, bool, None, Tuple["JSONValue"], List["JSONValue"], Dict[str, "JSONValue"]]
JSONData = Union[List[Dict[str, JSONValue]], Dict[str, JSONValue]]


class ClientMethodKwargs(TypedDict, total=False):
    content: RequestContent
    data: RequestData
    files: RequestFiles
    json: JSONData
    params: QueryParamTypes
    headers: HeaderTypes
    cookies: CookieTypes
    auth: Union[AuthTypes, UseClientDefault]
    follow_redirects: Union[bool, UseClientDefault]
    timeout: Union[TimeoutTypes, UseClientDefault]
    extensions: Mapping[str, Any]


class HooksData(TypedDict, total=False):
    CREATE: Union[str, Callable, None]
    UPDATE: Union[str, Callable, None]
    DELETE: Union[str, Callable, None]


class SignalChoices(models.IntegerChoices):
    CREATE = 0, "CREATE"
    UPDATE = 1, "UPDATE"
    DELETE = 2, "DELETE"
    CREATE_OR_UPDATE = 3, "CREATE OR UPDATE"
    CREATE_OR_DELETE = 4, "CREATE OR DELETE"
    UPDATE_OR_DELETE = 5, "UPDATE OR DELETE"
    ALL = 6, "ALL"

    @classmethod
    def create_choises(cls) -> Set["SignalChoices"]:
        return {
            cls.CREATE,
            cls.CREATE_OR_UPDATE,
            cls.CREATE_OR_DELETE,
            cls.ALL,
        }

    @classmethod
    def update_choises(cls) -> Set["SignalChoices"]:
        return {
            cls.UPDATE,
            cls.CREATE_OR_UPDATE,
            cls.UPDATE_OR_DELETE,
            cls.ALL,
        }

    @classmethod
    def delete_choises(cls) -> Set["SignalChoices"]:
        return {
            cls.DELETE,
            cls.CREATE_OR_DELETE,
            cls.UPDATE_OR_DELETE,
            cls.ALL,
        }
